"""Manages the configuration of PhysBiblio, from saving to reading.

This file is part of the physbiblio package.
"""
import ast
import logging
import logging.handlers
import os
import sys
import traceback
from collections import OrderedDict

from appdirs import AppDirs

try:
    from physbiblio.databaseCore import PhysBiblioDBCore, PhysBiblioDBSub
    from physbiblio.strings.main import ConfigStrings as cstr
    from physbiblio.tablesDef import profilesSettingsTable, searchesTable, tableFields
except ImportError:
    print("Could not find physbiblio and its modules!")
    print(traceback.format_exc())
    raise


globalLogName = "physbibliolog"
FNFError = FileNotFoundError


class ConfigParameter:
    """Instances contain some properties for configuration parameters,
    such as name, description, special type and default value
    """

    def __init__(self, name, default, description="", special=None, isGlobal=False):
        """Store the passed properties

        Parameters:
            name: the parameter name in the database
            default: the default value if not edited by the user
            description (default ""): a description
                for the configuration dialog
            special (default None): define a special type
                for the parameter (treat as list, int, float, bool, ...)
            isGlobal (default False): if True, the parameter is global
                and shared among all the databases.
                If False, it is relative to the current database only
        """
        self.name = name
        self.default = default
        self.description = description
        self.special = special
        self.isGlobal = isGlobal

    def __getitem__(self, key):
        """Extract attribute using a [...] syntax,
        for compatibility with previous code
        """
        try:
            item = getattr(self, key)
        except AttributeError:
            item = None
            logging.getLogger(globalLogName).warning(cstr.noAttribute % key)
        return item


class ConfigParametersList(OrderedDict):
    """Inherit from OrderedDict, but add a quick method to add
    `ConfigParameter`s to the list"""

    def add(self, cp):
        """Add a new value, corresponding to the key that is cp.name

        Parameters:
            cp: a ConfigParameter instance
        """
        if isinstance(cp, ConfigParameter):
            self[cp.name] = cp
        else:
            logging.getLogger(globalLogName).warning(cstr.invalidType % type(cp))


# these are the default parameter values, descriptions and types
configuration_params = ConfigParametersList()
configuration_params.add(
    ConfigParameter(
        "mainDatabaseName",
        "PBDATAphysbiblio.db",
        description=cstr.Desc.mainDBName,
    )
)
configuration_params.add(
    ConfigParameter(
        "loggingLevel",
        1,
        description=cstr.Desc.logLevel,
        special="int",
    )
)
configuration_params.add(
    ConfigParameter(
        "logFileName",
        "PBDATAparams.log",
        description=cstr.Desc.logFName,
    )
)
configuration_params.add(
    ConfigParameter(
        "pdfFolder",
        "PBDATApdf",
        description=cstr.Desc.PDFFolder,
    )
)
configuration_params.add(
    ConfigParameter(
        "pdfApplication",
        "",
        description=cstr.Desc.PDFApp,
    )
)
configuration_params.add(
    ConfigParameter(
        "webApplication",
        "",
        description=cstr.Desc.webApp,
    )
)
configuration_params.add(
    ConfigParameter(
        "timeoutWebSearch",
        20.0,
        description=cstr.Desc.timeout,
        special="float",
    )
)
configuration_params.add(
    ConfigParameter(
        "askBeforeExit", False, description=cstr.Desc.confirmExit, special="boolean"
    )
)
configuration_params.add(
    ConfigParameter(
        "defaultLimitBibtexs",
        100,
        description=cstr.Desc.limitBibtexs,
        special="int",
    )
)
configuration_params.add(
    ConfigParameter(
        "defaultUpdateFrom",
        0,
        description=cstr.Desc.updateFrom,
        special="int",
    )
)
configuration_params.add(
    ConfigParameter(
        "maxAuthorNames",
        3,
        description=cstr.Desc.maxAuthorsD,
        special="int",
    )
)
configuration_params.add(
    ConfigParameter(
        "maxAuthorSave",
        6,
        description=cstr.Desc.maxAuthorsS,
        special="int",
    )
)
configuration_params.add(
    ConfigParameter(
        "maxExternalAPIResults",
        10,
        description=cstr.Desc.maxAPIRes,
        special="int",
    )
)
configuration_params.add(
    ConfigParameter(
        "fetchAbstract",
        False,
        description=cstr.Desc.fetchAbstract,
        special="boolean",
    )
)
configuration_params.add(
    ConfigParameter(
        "defaultCategories",
        [],
        description=cstr.Desc.defaultCat,
        special="list",
    )
)
configuration_params.add(
    ConfigParameter(
        "bibListFontSize",
        9,
        description=cstr.Desc.fontSize,
        special="float",
    )
)
configuration_params.add(
    ConfigParameter(
        "bibtexListColumns",
        [
            "bibkey",
            "author",
            "title",
            "year",
            "firstdate",
            "pubdate",
            "doi",
            "arxiv",
            "isbn",
            "inspire",
        ],
        description=cstr.Desc.bibListCols,
        special="list",
    )
)
configuration_params.add(
    ConfigParameter(
        "resizeTable",
        True,
        description=cstr.Desc.autoResize,
        special="boolean",
    )
)
configuration_params.add(
    ConfigParameter(
        "maxSavedSearches",
        5,
        description=cstr.Desc.maxSavedSearches,
        special="int",
    )
)
configuration_params.add(
    ConfigParameter(
        "ADSToken",
        "",
        description=cstr.Desc.ADSToken,
    )
)
configuration_params.add(
    ConfigParameter(
        "openSinceLastUpdate",
        "0.0.0",
        description=cstr.Desc.sinceLastUpdate,
        isGlobal=True,
    )
)
configuration_params.add(
    ConfigParameter(
        "notifyUpdate",
        True,
        description=cstr.Desc.notifyUpdate,
        special="boolean",
        isGlobal=True,
    )
)
configuration_params.add(
    ConfigParameter(
        "batchSizeInspire",
        50,
        description=cstr.Desc.limitBibtexs,
        special="int",
    )
)

ignoreParameterOrder = ["mainDatabaseName", "openSinceLastUpdate"]
loggingLevels = ["0 - errors", "1 - warnings", "2 - info", "3 - all"]


class GlobalDB(PhysBiblioDBCore):
    """Class that manages the operations on the global DB:
    profiles and frequent searches/replaces
    """

    def __init__(self, dbname, logger, datapath, info=True):
        """Class constructor

        Parameters:
            dbname: the name of the database to open
            logger: the logger used for messages
            datapath: the path where to store database files
            info (boolean, default True): print some output messages
        """
        self.dataPath = datapath
        PhysBiblioDBCore.__init__(self, dbname, logger, noOpen=True)
        self.openDB(info=info)

        self.cursExec("SELECT name FROM sqlite_master WHERE type='table';")
        tables = [name[0] for name in self.curs]
        if not all([a in tables for a in ("profiles", "searches", "settings")]):
            self.createTables(tables)

        if self.countProfiles() == 0:
            self.createProfile()
        self.checkOnlyFilename()

        self.config = ConfigurationDB(self)

    def createTables(self, existing):
        """Create the profiles table"""
        if "profiles" not in existing:
            self.createTable("profiles", profilesSettingsTable, critical=True)
        if "searches" not in existing:
            self.createTable("searches", searchesTable, critical=True)
        if "settings" not in existing:
            self.createTable("settings", tableFields["settings"], critical=True)
        return True

    def countProfiles(self):
        """Obtain the number of profiles in the database

        Output:
            the number of profiles
        """
        self.cursExec("SELECT Count(*) FROM profiles")
        return self.curs.fetchall()[0][0]

    def createProfile(
        self, name="default", description="", databasefile=None, oldCfg=""
    ):
        """Create a new profile

        Parameters:
            name: the name of the profile
            description: a short description of the profile
            databasefile: the name of the database file
                which will contain data
            oldCfg: the name of the old .cfg file
                (backwards compatibility only)

        Output:
            True if successful, sys.exit(1) otherwise
        """
        if databasefile is None:
            databasefile = (
                configuration_params["mainDatabaseName"]
                .default.replace("PBDATA", "")
                .replace(self.dataPath, "")
            )
        command = (
            "INSERT into profiles "
            + "(name, description, databasefile, oldCfg, isDefault, ord) "
            + "values (:name, :description, :databasefile, "
            + ":oldCfg, :isDefault, :ord)"
        )
        data = {
            "name": name,
            "description": description,
            "databasefile": databasefile,
            "oldCfg": oldCfg,
            "isDefault": 1 if self.countProfiles() == 0 else 0,
            "ord": 100,
        }
        self.logger.info("%s\n%s" % (command, data))
        if not self.connExec(command, data):
            self.logger.exception(cstr.errorInsertProfile)
            sys.exit(1)
        self.commit(verbose=False)
        return True

    def updateProfileField(self, identifier, field, value, identifierField="name"):
        """Update a field of an existing profile

        Parameters:
            identifier: the identifier of the profile.
                It may be 1 if `identifierField` == "isDefault",
                the name of the profile or the database filename
                if `identifierField` == "name" or "databasefile",
                respectively.
            field: the name of the field to update.
            value: the new vallue of the field
            identifierField: "name", "databasefile" or "isDefault".
                It must be "databasefile" if you want to change
                the profile name and "name" if you want to change
                the databasefile (discouraged, however)

        Output:
            True if successful, False otherwise
        """
        if (
            (field == "databasefile" and identifierField != "name")
            or (field == "name" and identifierField != "databasefile")
            or (identifierField == "isDefault" and identifier != 1)
            or identifierField not in ("name", "databasefile", "isDefault")
            or field not in [e[0] for e in profilesSettingsTable]
        ):
            self.logger.error(cstr.errorInvalidFieldI % (field, identifierField))
            return False
        command = (
            "update profiles set %s = :val " % field
            + " where %s = :iden\n" % identifierField
        )
        data = {"val": value, "iden": identifier}
        self.logger.debug("%s\n%s" % (command, data))
        if not self.connExec(command, data):
            self.logger.error(cstr.errorUpdateProfile)
            return False
        self.commit(verbose=False)
        return True

    def deleteProfile(self, name):
        """Delete a profile given the name

        Parameters:
            name: the profile name

        Output:
            True if successful, False otherwise
        """
        if name.strip() == "":
            self.logger.error(cstr.errorProfileName)
            return False
        isDefault = name == self.getDefaultProfile()
        command = "delete from profiles where name = :name\n"
        self.logger.debug("%s\n%s" % (command, {"name": name}))
        if not self.connExec(command, {"name": name}):
            self.logger.error(cstr.errorDeleteProfile)
            return False
        if isDefault:
            self.getDefaultProfile()
        self.commit(verbose=False)
        return True

    def getProfiles(self):
        """Get the list of profiles

        Output:
            the list of `sqlite3.Row` objects
        """
        self.cursExec("SELECT * FROM profiles order by ord ASC, name ASC\n")
        return self.curs.fetchall()

    def getProfile(self, name="", filename=""):
        """Get a profile given its name or the name of the database file.
        Note that only one of the two may be given

        Parameters:
            name: the name of the profile
            filename: the database filename of the profile

        Output:
            the `sqlite3.Row` object
        """
        if name.strip() == filename.strip() == "":
            self.logger.warning(cstr.errorNameFilenameOne)
            return {}
        if name.strip() != "" and filename.strip() != "":
            self.logger.warning(cstr.errorNameFilenameOnly)
            return {}
        self.cursExec(
            "SELECT * FROM profiles WHERE name = :name or databasefile = :file\n",
            {"name": name, "file": filename},
        )
        return self.curs.fetchall()[0]

    def getProfileOrder(self):
        """Obtain the order of profiles

        Output:
            a list with the names of the ordered profiles
        """
        if self.countProfiles() == 0:
            self.createProfile()
            self.setDefaultProfile("default")
        self.cursExec("SELECT * FROM profiles order by ord ASC, name ASC\n")
        return [e["name"] for e in self.curs.fetchall()]

    def setProfileOrder(self, order=[]):
        """Set the new order of profiles

        Parameters:
            order: the ordered list of profile names

        Output:
            True if successful, False otherwise
        """
        if order == []:
            self.logger.warning(cstr.errorOrder)
            return False
        if sorted(order) != sorted([e["name"] for e in self.getProfiles()]):
            self.logger.info(sorted(order))
            self.logger.info(sorted([e["name"] for e in self.getProfiles()]))
            self.logger.warning(cstr.errorListProfilesMatch)
            return False
        failed = False
        for ix, profName in enumerate(order):
            if not self.connExec(
                "update profiles set ord=:ord where name=:name\n",
                {"name": profName, "ord": ix},
            ):
                failed = True
        if not failed:
            self.commit(verbose=False)
        else:
            self.logger.error(cstr.errorProfileOrdering)
            self.undo(verbose=False)
        return not failed

    def getDefaultProfile(self):
        """Obtain the name of the default profile

        Output:
            the `sqlite3.Row` objects
        """
        if self.countProfiles() == 0:
            self.createProfile()
            self.setDefaultProfile("default")
        self.cursExec("SELECT * FROM profiles WHERE isDefault = 1\n")
        try:
            defaultProfileName = [e["name"] for e in self.curs.fetchall()][0]
        except IndexError:
            self.cursExec("SELECT * FROM profiles\n")
            defaultProfileName = [e["name"] for e in self.curs.fetchall()][0]
            if self.setDefaultProfile(defaultProfileName):
                self.logger.info(cstr.defaultProfileChanged % defaultProfileName)
        return defaultProfileName

    def setDefaultProfile(self, name=None):
        """Set the new default profile

        Parameters:
            name: the profile name

        Output:
            True if successful, False otherwise
        """
        if name is None or name.strip() == "":
            self.logger.warning(cstr.errorName)
            return False
        self.cursExec("SELECT * FROM profiles WHERE name = :name\n", {"name": name})
        if len(self.curs.fetchall()) == 1:
            if self.connExec(
                "update profiles set isDefault=0 where 1\n"
            ) and self.connExec(
                "update profiles set isDefault=1 where name = :name\n", {"name": name}
            ):
                self.commit(verbose=False)
                return True
            else:
                self.logger.error(cstr.errorProfileDefault)
                self.undo(verbose=False)
                return False
        else:
            self.logger.warning(cstr.errorNoProfilesName)
            return False

    def checkOnlyFilename(self):
        """Convert previously stored profile names"""
        for q in self.getProfiles():
            if q["databasefile"].startswith(self.dataPath):
                self.updateProfileField(
                    q["name"],
                    "databasefile",
                    q["databasefile"].replace(self.dataPath, ""),
                )

    def countSearches(self):
        """Obtain the number of searches in the table

        Output:
            the number of searches
        """
        self.cursExec("SELECT Count(*) FROM searches")
        return self.curs.fetchall()[0][0]

    def insertSearch(
        self,
        name="",
        count=0,
        searchFields=[],
        replaceFields=[],
        manual=False,
        replacement=False,
        limit=0,
        offset=0,
    ):
        """Insert a new search/replace

        Parameters:
            name: the config name
            count: the order in the cronology or in the menu
            searchFields: the list of dictionaries which is meant
                to be passed to fetchByDict
            replaceFields: the replace fields used in searchAndReplace
            manual (boolean, default False): manually saved entry
            replacement (boolean, default False): replace or simple search
            limit: the number of requested entries
            offset: the offset

        Output:
            the output of self.connExec
        """
        if limit == 0:
            limit = pbConfig.params["defaultLimitBibtexs"]
        output = self.connExec(
            "INSERT into searches "
            + "(name, count, searchDict, limitNum, offsetNum, replaceFields,"
            + " manual, isReplace) values (:name, :count, :searchFields,"
            + " :limit, :offset, :replaceFields, :manual, :isReplace)\n",
            {
                "name": name,
                "count": count,
                "searchFields": "%s" % searchFields,
                "limit": limit,
                "offset": offset,
                "replaceFields": "%s" % replaceFields,
                "manual": 1 if manual else 0,
                "isReplace": 1 if replacement else 0,
            },
        )
        self.commit()
        return output

    def deleteSearch(self, idS):
        """Delete a search/replace given the id.

        Parameters:
            idS: the unique identifier

        Output:
            the output of cursExec
        """
        output = self.cursExec("delete from searches where idS=?\n", (idS,))
        self.commit()
        return output

    def getAllSearches(self):
        """Get all the searches

        Output:
            the list of `sqlite3.Row` objects
                with all the searches in the database
        """
        self.cursExec("select * from searches order by count asc\n")
        return self.curs.fetchall()

    def getSearchByID(self, idS):
        """Get a search given its name

        Parameters:
            idS: the name of the search/replace

        Output:
            the list (len = 1) of `sqlite3.Row` objects
                with all the matching searches
        """
        self.cursExec("select * from searches where idS=?\n", (idS,))
        return self.curs.fetchall()

    def getSearchByName(self, name):
        """Get a search given its name

        Parameters:
            name: the name of the search/replace

        Output:
            the list of `sqlite3.Row` objects with all the matching searches
        """
        self.cursExec("select * from searches where name=?\n", (name,))
        return self.curs.fetchall()

    def getSearchList(self, manual=False, replacement=False):
        """Get searches or replaces which were not manually saved

        Parameters:
            manual (boolean, default False): manually saved entry
            replacement (boolean, default False): replace or simple search

        Output:
            the list of `sqlite3.Row` objects with all the matching searches
        """
        self.cursExec(
            "select * from searches where manual=? "
            + "and isReplace=? order by count ASC\n",
            (1 if manual else 0, 1 if replacement else 0),
        )
        return self.curs.fetchall()

    def updateSearchOrder(self, replacement=False):
        """Update the cronology order for searches or replaces
        which were not manually saved

        Parameters:
            replacement (boolean, default False): replace or simple search

        Output:
            True if successfull, False if some sum failed
        """
        self.cursExec(
            "select * from searches where manual=? and isReplace=?\n",
            (0, 1 if replacement else 0),
        )
        for e in self.curs.fetchall():
            if e["count"] + 1 >= pbConfig.params["maxSavedSearches"]:
                self.deleteSearch(e["idS"])
            elif not self.connExec(
                "update searches set count = :count where idS=:idS\n",
                {"idS": e["idS"], "count": e["count"] + 1},
            ):
                self.undo()
                return False
        self.commit()
        return True

    def updateSearchField(self, idS, field, value, isReplace=True):
        """Update a single field of an entry

        Parameters:
            idS: the search ID
            field: the field name
            value: the new value of the field

        Output:
            the output of self.connExec or
                False (empty value or not valid field)
        """
        if (field == "replaceFields" and not isReplace) or (
            field in ("searchDict", "replaceFields", "name", "limitNum", "offsetNum")
            and value is not None
            and ("%s" % value).strip() not in ("", "[]", "{}")
        ):
            query = "update searches set " + field + "=:field where idS=:idS\n"
            return self.connExec(query, {"field": "%s" % value, "idS": idS})
        else:
            self.logger.warning(cstr.errorSearchField % (field, value))
            return False


class ConfigurationDB(PhysBiblioDBSub):
    """Subclass that manages the functions for the categories."""

    def count(self):
        """Obtain the number of settings in the table

        Output:
            the number of settings
        """
        self.cursExec("SELECT Count(*) FROM settings")
        return self.curs.fetchall()[0][0]

    def insert(self, name, value):
        """Insert a new setting. If already existing, update it

        Parameters:
            name: the config name
            value: the content of the setting

        Output:
            the output of self.connExec
        """
        self.cursExec("select * from settings where name=?\n", (name,))
        if len(self.curs.fetchall()) > 0:
            self.mainDB.logger.info(cstr.confEntryUpdate)
            return self.update(name, value)
        return self.connExec(
            "INSERT into settings (name, value) values (:name, :value)\n",
            {"name": name, "value": value},
        )

    def update(self, name, value):
        """Update an existing setting

        Parameters:
            name: the config name
            value: the content of the setting

        Output:
            the output of self.connExec
        """
        self.cursExec("select * from settings where name=?\n", (name,))
        if len(self.curs.fetchall()) == 0:
            self.mainDB.logger.info(cstr.confEntryInsert % name)
            return self.insert(name, value)
        return self.connExec(
            "update settings set value = :value where name = :name\n",
            {"name": name, "value": value},
        )

    def delete(self, name):
        """Delete a setting.

        Parameters:
            name: the setting name

        Output:
            the output of cursExec
        """
        return self.cursExec("delete from settings where name=?\n", (name,))

    def getAll(self):
        """Get all the settings

        Output:
            the list of `sqlite3.Row` objects with
                all the settings in the database
        """
        self.cursExec("select * from settings\n")
        return self.curs.fetchall()

    def getByName(self, name):
        """Get a setting given its name

        Parameters:
            name: the name of the setting

        Output:
            the list (len = 1) of `sqlite3.Row` objects
            with all the matching settings
        """
        self.cursExec("select * from settings where name=?\n", (name,))
        return self.curs.fetchall()


def getLogLevel(value):
    """Convert the numerical code saved in the configuration
    to a logging level code

    Parameter:
        value: an int, the logging level in the PhysBiblio scale

    Output:
        an int (may be among
            logging.ERROR, logging.WARNING, logging.INFO, logging.DEBUG
        )
    """
    if value == 0:
        return logging.ERROR
    elif value == 1:
        return logging.WARNING
    elif value == 2:
        return logging.INFO
    return logging.DEBUG


def replacePBDATA(path, var):
    """Replace the PBDATA placeholder in the given variable,
    if it is a string, or just return the variable otherwise

    Parameters:
        var: the input variable

    Output:
        the input variable or a string with replaced "PBDATA"
    """
    if isinstance(var, str) and "PBDATA" in var:
        var = os.path.join(path, var.replace("PBDATA", ""))
    return var


def addFileHandler(logger, logFileName, defaultPath=""):
    """Create and add a RotatingFileHandler to an existing Logger

    Parameters:
        logger: an existing instance of logging.Logger
        logFileName: the name of the log file
        defaultPath (default ""): used in case the logFileName is missing
            to point to the default log file name in the default folder
    """
    try:
        fh = logging.handlers.RotatingFileHandler(
            logFileName, maxBytes=5.0 * 2**20, backupCount=5
        )
    except FNFError:
        logger.exception("")
        fh = logging.handlers.RotatingFileHandler(
            replacePBDATA(defaultPath, configuration_params["logFileName"].default),
            maxBytes=5.0 * 2**20,
            backupCount=5,
        )
    fh.setLevel(getLogLevel(pbConfig.params["loggingLevel"]))
    formatter = logging.Formatter(
        "%(asctime)s %(levelname)10s : [%(module)s.%(funcName)s] %(message)s"
    )
    fh.setFormatter(formatter)
    logger.addHandler(fh)


class ConfigVars:
    """Contains all the common settings, the information on the profiles
    and their configuration.
    """

    adsUrl = "https://ui.adsabs.harvard.edu/abs/"
    arxivUrl = "https://arxiv.org"
    doiUrl = "https://doi.org/"
    inspireAPI = "https://inspirehep.net/api/"
    inspireConferencesAPI = "https://inspirehep.net/api/conferences/"
    inspireExperimentsLink = "https://inspirehep.net/experiments/"
    inspireLiteratureAPI = "https://inspirehep.net/api/literature/"
    inspireLiteratureLink = "https://inspirehep.net/literature/"

    def __init__(self, profileFileName="profiles.db"):
        """Initialize the configuration.
        Check the profiles first, then load the default profile
        and its configuration.
        """
        self.defaultProfileName = None
        self.profiles = None
        self.profileOrder = None
        self.currentProfileName = None
        self.currentProfile = None
        self.currentDatabase = ""
        self.loggerString = None
        self.logger = None

        # needed because the main logger will be loaded later!
        logging.basicConfig(
            format="[%(module)s.%(funcName)s] %(message)s", level=logging.INFO
        )
        self.prepareLogger(globalLogName)
        self.defaultDirs = AppDirs("PhysBiblio")

        configPath = os.getenv(
            "PHYSBIBLIO_CONFIG", self.defaultDirs.user_config_dir + os.sep
        )
        self.configPath = (
            configPath
            if isinstance(configPath, str) and configPath.strip() != ""
            else self.defaultDirs.user_config_dir + os.sep
        )
        dataPath = os.getenv("PHYSBIBLIO_DATA", self.defaultDirs.user_data_dir + os.sep)
        self.dataPath = (
            dataPath
            if isinstance(dataPath, str) and dataPath.strip() != ""
            else self.defaultDirs.user_data_dir + os.sep
        )
        self.logger.info(cstr.defaultCfgPath % self.configPath)
        if not os.path.exists(self.configPath):
            os.makedirs(self.configPath)
        self.logger.info(cstr.defaultDataPath % self.dataPath)
        if not os.path.exists(self.dataPath):
            os.makedirs(self.dataPath)

        self.loggingLevels = loggingLevels
        self.paramOrder = [
            p.name
            for p in configuration_params.values()
            if p.name not in ignoreParameterOrder
        ]
        self.setDefaultParams()

        self.oldConfigProfilesFile = os.path.join(self.configPath, "profiles.dat")
        self.globalDbFile = os.path.join(self.configPath, profileFileName)
        self.globalDb = GlobalDB(
            self.globalDbFile, self.logger, self.dataPath, info=False
        )

        self.checkOldProfiles()
        self.loadProfiles()

    def loadProfiles(self):
        """Load the information from the profile database"""
        try:
            (
                self.defaultProfileName,
                self.profiles,
                self.profileOrder,
            ) = self.readProfiles()
        except (IOError, ValueError, SyntaxError) as e:
            self.logger.warning(e)
            self.globalDb.createProfile()

    def prepareLogger(self, string):
        """Replace the logger used by this module

        Parameters:
            string: the string used in getLogger
        """
        self.loggerString = string
        self.logger = logging.getLogger(self.loggerString)

    def readConfig(self):
        """Read the configuration from the current database.
        Single parameters are read by self.readParam
        """
        self.logger.debug(cstr.readConf)
        self.setDefaultParams()
        tempDb = PhysBiblioDBCore(self.currentDatabase, self.logger, info=False)
        configDb = ConfigurationDB(tempDb)
        try:
            for k in configuration_params.keys():
                self.readParam(k, configDb)
        except Exception:
            self.logger.exception(cstr.errorReadConf % (self.currentDatabase))
        tempDb.closeDB(info=False)
        rfhs = []
        for lh in self.logger.handlers:
            if isinstance(lh, logging.handlers.RotatingFileHandler):
                rfhs.append(lh)
        for lh in rfhs:
            self.logger.removeHandler(lh)
        addFileHandler(
            self.logger, self.params["logFileName"], defaultPath=self.dataPath
        )
        self.logger.debug(cstr.confLoaded)

    def readParam(self, key, configDb):
        """Read the value of a single parameter from the database,
        and process it according to the parameter type.
        The result is stored in `self.params[key]`

        Parameters:
            key: the parameter name
            configDb: the configuration database where to read from
        """
        if key == "mainDatabaseName":
            return
        if configuration_params[key].isGlobal:
            cont = self.globalDb.config.getByName(key)
        else:
            cont = configDb.getByName(key)
        if len(cont) == 0:
            self.params[key] = self.replacePBDATA(configuration_params[key].default)
            return
        v = cont[0]["value"]
        try:
            if configuration_params[key].special == "float":
                self.params[key] = float(v)
            elif configuration_params[key].special == "int":
                self.params[key] = int(v)
            elif configuration_params[key].special == "boolean":
                if v.lower() in ("true", "1", "yes", "on"):
                    self.params[key] = True
                elif v.lower() in ("false", "0", "no", "off"):
                    self.params[key] = False
                else:
                    raise ValueError
            elif configuration_params[key].special == "list":
                self.params[key] = ast.literal_eval(v.strip())
            else:
                self.params[key] = self.replacePBDATA(v)
        except Exception:
            self.logger.warning(cstr.errorReadParam % key, exc_info=True)
            self.params[key] = configuration_params[key].default

    def readProfiles(self):
        """Reads the list of profiles and the related parameters
        from the profiles database.

        Output:
            the name of the default profile, the dictionary
                with the profiles and the list of ordered profile names
        """
        allProf = self.globalDb.getProfiles()
        profiles = {}
        for e in allProf:
            profiles[e["name"]] = {
                "n": e["name"],
                "d": e["description"],
                "f": e["oldCfg"],
                "db": e["databasefile"]
                if os.sep in e["databasefile"]
                else os.path.join(self.dataPath, e["databasefile"]),
            }
        return (
            self.globalDb.getDefaultProfile(),
            profiles,
            self.globalDb.getProfileOrder(),
        )

    def reInit(self, newShort, newProfile=None):
        """Used when changing profile.
        Reload all the configuration given the new profile name.

        Parameters:
            newShort (str): short name for the new profile to be loaded
            newProfile (dict, optional): the profile file dictionary
        """
        if newProfile is None:
            try:
                newProfile = self.profiles[newShort]
            except KeyError:
                self.logger.error(cstr.errorProfileNotFound)
                return
        self.currentProfileName = newShort
        self.currentProfile = newProfile
        self.currentDatabase = (
            newProfile["db"]
            if os.sep in newProfile["db"]
            else os.path.join(self.dataPath, newProfile["db"])
        )
        self.setDefaultParams()
        self.logger.info(
            cstr.restartWProfile % (self.currentProfileName, self.currentDatabase)
        )
        self.readConfig()

    def reloadProfiles(self, useProfile=None):
        """Load the information from the profile database,
        reset the default and current profile and the settings

        Parameters:
            useProfile (optional): the name of the profile
                to be used instead of the default one
        """
        self.loadProfiles()

        self.currentProfileName = (
            self.defaultProfileName if useProfile is None else useProfile
        )
        try:
            self.currentProfile = self.profiles[self.currentProfileName]
        except KeyError:
            self.logger.critical(
                cstr.errorNoProfileUseDef % (useProfile, self.defaultProfileName)
            )
            self.currentProfileName = self.defaultProfileName
            self.currentProfile = self.profiles[self.currentProfileName]
        self.currentDatabase = (
            self.currentProfile["db"]
            if os.sep in self.currentProfile["db"]
            else os.path.join(self.dataPath, self.currentProfile["db"])
        )
        self.logger.info(
            cstr.startWProfile % (self.currentProfileName, self.currentDatabase)
        )

        self.readConfig()

    def replacePBDATA(self, var):
        """Replace the PBDATA placeholder in the given variable,
        if it is a string, or just return the variable otherwise

        Parameters:
            var: the input variable

        Output:
            the input variable or a string with replaced "PBDATA"
        """
        return replacePBDATA(self.dataPath, var)

    def setDefaultParams(self):
        """Reset the list of params,
        setting their value to the default
        as defined in configuration_params
        """
        self.params = {}
        for k, p in configuration_params.items():
            if k == "mainDatabaseName":
                continue
            self.params[k] = self.replacePBDATA(p.default)

    def checkOldProfiles(self):
        """Intended for backwards compatibility.
        Check if there is a profiles.dat file in the self.configPath folder.
        If yes, move it and the related information into the databases.
        """
        if os.path.isfile(self.oldConfigProfilesFile):
            self.logger.info("Moving info from profiles.dat into the profiles.db")
            for k in self.globalDb.getProfileOrder():
                self.globalDb.deleteProfile(k)
            defProf, profiles, profileOrder = self.oldReadProfiles()
            for k in profiles.keys():
                self.oldReInit(k, profiles[k])
                tempDb = PhysBiblioDBCore(self.params["mainDatabaseName"], self.logger)
                configDb = ConfigurationDB(tempDb)

                self.globalDb.createProfile(
                    name=k,
                    description=profiles[k]["d"],
                    databasefile=self.params["mainDatabaseName"],
                    oldCfg=profiles[k]["f"],
                )
                if k == defProf:
                    self.globalDb.setDefaultProfile(k)
                for p in self.paramOrder:
                    if self.params[p] != configuration_params[p].default:
                        configDb.insert(p, str(self.params[p]))
                tempDb.commit()
                tempDb.closeDB()
                oldfile = os.path.join(self.configPath, profiles[k]["f"])
                os.rename(oldfile, oldfile + "_bck")
                self.logger.info(
                    "Old '%s' renamed to '%s'." % (oldfile, oldfile + "_bck")
                )
            self.globalDb.setProfileOrder(profileOrder)
            self.logger.info([dict(e) for e in self.globalDb.getProfiles()])
            os.rename(self.oldConfigProfilesFile, self.oldConfigProfilesFile + "_bck")
            self.logger.info(
                "Old '%s' renamed to '%s'."
                % (self.oldConfigProfilesFile, self.configProfilesFile + "_bck")
            )

    def oldReadConfigFile(self):
        """Read the configuration from a file,
        whose name is stored in self.configMainFile.
        Parses the various parameters given their declared type.
        """
        for k, p in configuration_params.items():
            if isinstance(p.default, str) and "PBDATA" in p.default:
                p.default = os.path.join(self.dataPath, p.default.replace("PBDATA", ""))
            self.params[k] = p.default
        try:
            with open(self.configMainFile) as r:
                txt = r.readlines()
            for l in txt:
                k, v = l.replace("\n", "").split(" = ")
                try:
                    if configuration_params[k].special == "float":
                        self.params[k] = float(v)
                    elif configuration_params[k].special == "int":
                        self.params[k] = int(v)
                    elif configuration_params[k].special == "boolean":
                        if v.lower() in ("true", "1", "yes", "on"):
                            self.params[k] = True
                        elif v.lower() in ("false", "0", "no", "off"):
                            self.params[k] = False
                        else:
                            raise ValueError
                    elif configuration_params[k].special == "list":
                        self.params[k] = ast.literal_eval(v.strip())
                    else:
                        if isinstance(v, str) and "PBDATA" in v:
                            v = os.path.join(self.dataPath, v.replace("PBDATA", ""))
                        self.params[k] = v
                except Exception:
                    self.logger.warning("Failed in reading parameter", exc_info=True)
                    self.params[k] = v
        except IOError:
            self.logger.warning(
                "ERROR: config file %s do not exist." % self.configMainFile
            )
        except Exception:
            self.logger.error("ERROR: reading %s file failed." % self.configMainFile)

    def oldReadProfiles(self):
        """Reads the list of profiles and the related parameters
        from the profiles.dat file.
        """
        with open(self.oldConfigProfilesFile) as r:
            txtarr = r.readlines()
        txt = "".join(txtarr)
        parsed = ast.literal_eval(txt.replace("\n", ""))
        if len(parsed) < 3:
            parsed = parsed + tuple(sorted(parsed[1].keys()))
        return parsed

    def oldReInit(self, newShort, newProfile):
        """Old function used when changing profile.
        Reloads all the configuration from scratch given the new profile name.

        Parameters:
            newShort (str): short name for the new profile to be loaded
            newProfile (dict): the profile file dictionary
        """
        self.currentProfileName = newShort
        self.params = {}
        for k, p in configuration_params.items():
            self.params[k] = p.default
        self.currentProfile = newProfile
        self.logger.info(
            "Starting with configuration in '%s'" % self.currentProfile["f"]
        )
        self.configMainFile = os.path.join(self.configPath, self.currentProfile["f"])
        self.params = {}
        self.oldReadConfigFile()


pbConfig = ConfigVars()



#!/usr/bin/env python
"""Test file for the physbiblio.inspireStats module.

This file is part of the physbiblio package.
"""
import datetime
import os
import traceback
import unittest
from io import StringIO
from unittest.mock import MagicMock, call, patch

import matplotlib
import pytz

try:
    from physbiblio.errors import pBErrorManager
    from physbiblio.inspireStats import *
    from physbiblio.setuptests import *
except ImportError:
    print("Could not find physbiblio and its modules!")
    raise
except Exception:
    print(traceback.format_exc())


@unittest.skipIf(skipTestsSettings.online, "Online tests")
@patch("logging.Logger.debug")
@patch("logging.Logger.info")
@patch("logging.Logger.warning")
@patch("logging.Logger.exception")
class TestInspireStatsMethods(unittest.TestCase):
    """Tests for methods in physbiblio.inspireStats"""

    def test_init(self, *args):
        """test some properties"""
        self.assertIsInstance(pBStats, InspireStatsLoader)
        tpBStats = InspireStatsLoader()
        self.assertEqual(tpBStats.authorPlotInfo, None)
        self.assertEqual(tpBStats.paperPlotInfo, None)

    def test_paperStats(self, *args):
        """Test paperStats function downloading real and fake data"""
        # not a valid record (it was cancelled):
        self.assertEqual(
            pBStats.paperStats("1358853"),
            {
                "aI": {},
                "citList": [
                    [
                        datetime.datetime.fromordinal(
                            datetime.date.today().toordinal()
                        ).replace(tzinfo=pytz.UTC)
                    ],
                    [0],
                ],
                "id": "1358853",
            },
        )

        # needs multiple pages to load all the content
        testGood = pBStats.paperStats("650592")
        self.assertTrue(testGood["id"], "650592")
        self.assertTrue(len(testGood["aI"]) > 950)
        self.assertEqual(len(testGood["aI"]) + 1, len(testGood["citList"][1]))
        self.assertEqual(len(testGood["citList"][0]), len(testGood["citList"][1]))
        self.assertFalse("fig" in testGood.keys())

        testGood = pBStats.paperStats("1385583", plot=True)
        self.assertTrue(testGood["id"], "1385583")
        self.assertTrue(len(testGood["aI"]) > 120)
        self.assertEqual(len(testGood["aI"]) + 1, len(testGood["citList"][1]))
        self.assertEqual(len(testGood["citList"][0]), len(testGood["citList"][1]))
        self.assertIsInstance(testGood["fig"], matplotlib.figure.Figure)

        testGood = pBStats.paperStats(["1358853", "1385583"])
        self.assertTrue(testGood["id"], ["1358853", "1385583"])
        self.assertTrue(len(testGood["aI"]) > 120)
        pbm = MagicMock()
        pbv = MagicMock()
        testGood = pBStats.paperStats(["1358853", "1385583"], pbMax=pbm, pbVal=pbv)
        self.assertTrue(testGood["id"], ["1358853", "1385583"])
        self.assertTrue(len(testGood["aI"]) > 120)
        pbm.assert_called_once_with(2)
        pbv.assert_has_calls([call(1), call(2)])

    def test_authorStats(self, *args):
        """Test paperStats function downloading real and fake data"""
        res = pBStats.authorStats("E.M.Zavanin.1")
        self.assertTrue(res["h"] >= 4)
        self.assertTrue(res["paLi"][1][-1] == 8)
        self.assertTrue(res["allLi"][1][-1] >= 320)
        self.assertTrue(res["meanLi"][1][-1] >= 40)
        with patch(
            "physbiblio.inspireStats.InspireStatsLoader.paperStats",
            return_value={
                "aI": {
                    1653464: {
                        "date": datetime.datetime(2018, 2, 7, 4, 24, 2).replace(
                            tzinfo=pytz.UTC
                        )
                    },
                    1662320: {
                        "date": datetime.datetime(2018, 3, 14, 4, 53, 45).replace(
                            tzinfo=pytz.UTC
                        )
                    },
                    1664547: {
                        "date": datetime.datetime(2018, 3, 29, 5, 8, 58).replace(
                            tzinfo=pytz.UTC
                        )
                    },
                    1663942: {
                        "date": datetime.datetime(2018, 3, 26, 3, 25, 34).replace(
                            tzinfo=pytz.UTC
                        )
                    },
                    1658582: {
                        "date": datetime.datetime(2018, 3, 6, 3, 58, 52).replace(
                            tzinfo=pytz.UTC
                        )
                    },
                },
                "id": "1649081",
                "citList": [
                    [
                        datetime.datetime(2018, 2, 7, 4, 24, 2).replace(
                            tzinfo=pytz.UTC
                        ),
                        datetime.datetime(2018, 3, 6, 3, 58, 52).replace(
                            tzinfo=pytz.UTC
                        ),
                        datetime.datetime(2018, 3, 14, 4, 53, 45).replace(
                            tzinfo=pytz.UTC
                        ),
                        datetime.datetime(2018, 3, 26, 3, 25, 34).replace(
                            tzinfo=pytz.UTC
                        ),
                        datetime.datetime(2018, 3, 29, 5, 8, 58).replace(
                            tzinfo=pytz.UTC
                        ),
                        datetime.date(2018, 4, 16),
                    ],
                    [1, 2, 3, 4, 5, 5],
                ],
            },
            autospec=True,
        ) as _mock:
            testGood = pBStats.authorStats("E.M.Zavanin.1", plot=True)
            for i in (
                "1229039",
                "1345462",
                "1366180",
                "1385583",
                "1387736",
                "1404176",
                "1460832",
                "1519902",
            ):
                self.assertIn(i, testGood["aI"].keys())
                self.assertIn(i, [a[0][1] for a in _mock.call_args_list])
            self.assertTrue(
                testGood.keys(), ["meanLi", "allLi", "aI", "paLi", "h", "name"]
            )
            self.assertTrue(testGood["h"], 4)
            self.assertTrue(testGood["name"], "E.M.Zavanin.1")
            self.assertEqual(len(testGood["paLi"][0]), len(testGood["aI"]))
            self.assertEqual(len(testGood["paLi"][1]), len(testGood["aI"]))
            self.assertEqual(len(testGood["allLi"][0]), 5 * len(testGood["aI"]))
            self.assertEqual(len(testGood["allLi"][1]), 5 * len(testGood["aI"]))
            self.assertEqual(len(testGood["meanLi"][0]), 5 * len(testGood["aI"]))
            self.assertEqual(len(testGood["meanLi"][1]), 5 * len(testGood["aI"]))
            self.assertEqual(len(testGood["figs"]), 6)
            for f in testGood["figs"]:
                self.assertIsInstance(f, matplotlib.figure.Figure)

            testGood = pBStats.authorStats(["A.Gallego.Cadavid.1", "E.M.Zavanin.1"])
            pbm = MagicMock()
            pbv = MagicMock()
            testGood = pBStats.authorStats(
                ["A.Gallego.Cadavid.1", "E.M.Zavanin.1"], pbMax=pbm, pbVal=pbv
            )
            pbm.assert_called_once_with(2)
            pbv.assert_has_calls([call(1), call(2)])

            testGood = pBStats.authorStats("E.M.Zavanin.1")
            pbm = MagicMock()
            pbv = MagicMock()
            testGood = pBStats.authorStats("E.M.Zavanin.1", pbMax=pbm, pbVal=pbv)
            pbm.assert_called_once_with(8)
            pbv.assert_has_calls([call(i + 1) for i in range(8)])


if __name__ == "__main__":
    unittest.main()



"""The package that contains the tests for the `physbiblio.gui` subpackage.

This file is part of the physbiblio package.
"""
from physbiblio import __author__, __email__, __version__, __version_date__

__all__ = [
    "test_basicDialogs",
    "test_bibWindows",
    "test_catWindows",
    "test_commonClasses",
    "test_dialogWindows",
    "test_errorManager",
    "test_expWindows",
    "test_inspireStatsGUI",
    "test_mainWindow",
    "test_marks",
    "test_profilesManager",
    "test_threadElements",
]



"""This module manages the saved PDF files and opens them if required.

This file is part of the physbiblio package.
"""
import os
import os.path as osp
import shutil
import subprocess
import traceback
from urllib.request import HTTPError, URLError, urlopen

try:
    from physbiblio.config import pbConfig
    from physbiblio.database import pBDB
    from physbiblio.errors import pBLogger
    from physbiblio.strings.main import PDFStrings as pstr
except ImportError:
    print("Could not find physbiblio and its modules!")
    print(traceback.format_exc())
    raise


class LocalPDF:
    """Class with functions to manage the PDF folder content
    and the stored material
    """

    pdfDir = pbConfig.params["pdfFolder"]
    pdfApp = pbConfig.params["pdfApplication"]
    badFNameCharacters = r'\/:*?"<>|' + "'"

    def __init__(self):
        """Init the class and set some default variables"""
        self.checkFolderExists()
        self.pdfDir = pbConfig.params["pdfFolder"]
        self.pdfApp = pbConfig.params["pdfApplication"]

    def checkFolderExists(self):
        """Check if the PDF folder exists. If not, create it"""
        if not os.path.isdir(self.pdfDir):
            pBLogger.info(pstr.folderMissing % self.pdfDir)
            os.makedirs(self.pdfDir)

    def badFName(self, filename):
        """Clean the filename substituting the bad characters with '_'

        Parameters:
            filename (string): the string containing the file name

        Output:
            the cleaned filename
        """
        newFilename = ""
        if not isinstance(filename, str):
            pBLogger.warning(pstr.wrongType % filename)
            return ""
        for i in filename:
            if i not in self.badFNameCharacters:
                newFilename += i
            else:
                newFilename += "_"
        return newFilename

    def getFileDir(self, key):
        """Obtain the name of the directory for a given entry.
        The name is cleaned and the absolute path is generated.

        Parameters:
            key (string): the bibtex key of the entry

        Output:
            the (cleaned) absolute path for the PDF files
                associated with the entry
        """
        return osp.join(self.pdfDir, self.badFName(key))

    def getFilePath(self, key, fileType):
        """Obtain the file path for a given file type (arxiv, doi, ...)
        of a given entry.
        It reads the field from the database in order
        to generate the file name.

        Parameters:
            key (string): the bibtex key of the entry
            fileType (string in
                ["arxiv", "inspire", "isbn", "doi", "ads", "scholar"]):
                basically the field in the database
                from which the name must be taken

        Output:
            the (cleaned) absolute path for the PDF file
        """
        if fileType not in ("arxiv", "inspire", "isbn", "doi", "ads", "scholar"):
            pBLogger.warning(pstr.errorField)
            return ""
        filename = pBDB.bibs.getField(key, fileType)
        if filename is None:
            pBLogger.warning(pstr.errorGetType % (fileType, key))
            return ""
        else:
            filename = self.badFName(filename)
            return osp.join(self.getFileDir(key), filename + ".pdf")

    def createFolder(self, key, noCheck=False):
        """Create the PDF folder for a given entry.

        Parameters:
            key (string): the bibtex key of the entry
            noCheck (bool): True to avoid checking
                if the directory exists. Default False
        """
        directory = self.getFileDir(key)
        if noCheck or not osp.exists(directory):
            os.makedirs(directory)

    def renameFolder(self, oldkey, newkey):
        """Rename the PDF folder for a given entry
        for which the bibtex key has been changed.

        Parameters:
            oldkey (string): the old bibtex key of the entry
            newkey (string): the new bibtex key of the entry
        """
        olddir = self.getFileDir(oldkey)
        if osp.exists(olddir):
            newdir = self.getFileDir(newkey)
            pBLogger.info(pstr.rename % (olddir, newdir))
            if shutil.copytree(olddir, newdir, dirs_exist_ok=True):
                return shutil.rmtree(olddir)
        return False

    def copyNewFile(self, key, origFileName, fileType=None, customName=None):
        """Copy a file in any place of the filesystem
        into the directory corresponding to the given entry

        Parameters:
            key (string): the bibtex key of the entry
            origFileName (string): the name of the file to be copied
            fileType (string in
                ["arxiv", "inspire", "isbn", "doi", "ads", "scholar"]):
                basically the field in the database
                from which the name must be taken.
                `customName` can be given instead.
            customName (string):
                a completely independent name for the file.
                It may be given instead of fileType.
        """
        if not osp.exists(self.getFileDir(key)):
            self.createFolder(key, True)
        if customName is not None:
            newFileName = osp.join(self.getFileDir(key), customName)
        elif fileType is not None:
            newFileName = self.getFilePath(key, fileType)
        else:
            pBLogger.warning(pstr.errorMissingArg)
            return False
        try:
            shutil.copy2(origFileName, newFileName)
            pBLogger.info(pstr.copied % (origFileName, newFileName))
            return True
        except:
            pBLogger.exception(pstr.errorCopy % (origFileName, newFileName))
            return False

    def copyToDir(self, outFolder, key, fileType=None, customName=None):
        """Copy a file from the directory corresponding
        to the given entry to the a directory in the filesystem.

        Parameters:
            outFolder (string): the name of the destination folder
            key (string): the bibtex key of the entry
            fileType (string in
                ["arxiv", "inspire", "isbn", "doi", "ads", "scholar"]):
                basically the field in the database from which
                the name must be taken. customName can be given instead.
            customName (string):
                a completely independent name for the file.
                It may be given instead of fileType.
        """
        if customName is not None:
            origFile = osp.join(self.getFileDir(key), customName)
        elif fileType is not None:
            origFile = self.getFilePath(key, fileType)
        else:
            pBLogger.warning(pstr.errorMissingArg)
            return False
        try:
            shutil.copy2(origFile, outFolder)
            pBLogger.info(pstr.copied % (origFile, outFolder))
            return True
        except:
            pBLogger.exception(pstr.errorCopy % (origFile, outFolder))
            return False

    def downloadArxiv(self, key, force=False):
        """Download the PDF file from arXiv for a given entry
        and save it in the proper folder.

        Parameters:
            key (string): the bibtex key of the entry
            force (boolean, optional, default False):
                force the replacement of the file if it already exists.
                If False and there is a PDF with the same name,
                no action is performed.
        """
        filename = self.getFilePath(key, "arxiv")
        if osp.exists(filename) and not force:
            pBLogger.info(pstr.pdfPresent)
            return True
        self.createFolder(key)
        url = pBDB.bibs.getArxivUrl(key, "pdf")
        if not url:
            pBLogger.warning(pstr.errorArxivUrl % key)
            return False
        pBLogger.info(pstr.downloading % url)
        try:
            response = urlopen(url)
        except (URLError, ConnectionError):
            pBLogger.exception(pstr.pdfNotFound % key + pstr.e404 % url)
            return False
        else:
            try:
                with open(filename, "wb") as newF:
                    newF.write(response.read())
            except OSError:
                pBLogger.exception(pstr.errorSave % filename)
            else:
                pBLogger.info(pstr.saved % filename)
                return os.path.exists(filename)

    def openFile(self, key, arg=None, fileType=None, fileNum=None, fileName=None):
        """Open a PDF file in an external application
        (if defined in the configuration).
        Does nothing if there is self.pdfApp is empty.

        Parameters:
            key (string): the bibtex key of the entry
            at least one of the following (they are considered in this order):
                arg: used to guess the fileType, fileNum or fileName
                fileType (string in
                    ["arxiv", "inspire", "isbn", "doi", "ads", "scholar"]):
                    basically the field in the database
                    from which the name must be taken
                fileNum: the number of the desired PDF
                    in the list of existing PDF files for the considered entry
                filaName: the specific file name
        """
        try:
            if arg in ("arxiv", "doi"):
                fileType = arg
            elif float(arg).is_integer():
                fileNum = int(arg)
        except:
            if arg is not None:
                fileName = arg
        try:
            if fileType is not None:
                fName = self.getFilePath(key, fileType)
            elif fileNum is not None:
                fName = self.getExisting(key, True)[fileNum]
            elif fileName is not None:
                fName = osp.join(self.getFileDir(key), fileName)
            else:
                pBLogger.warning(pstr.errorInvalidSel)
                return
            if self.pdfApp != "":
                pBLogger.info(pstr.opening % fName)
                subprocess.Popen(
                    [self.pdfApp, fName],
                    stdout=subprocess.PIPE,
                    stderr=subprocess.STDOUT,
                )
        except:
            pBLogger.exception(pstr.openingFailed % ("PDF", key))

    def checkFile(self, key, fileType):
        """Check if a file of a given type (arxiv, doi, ...)
        and a given entry exists.

        Parameters:
            key (string): the bibtex key of the entry
            fileType (string in
                ["arxiv", "inspire", "isbn", "doi", "ads", "scholar"]):
                basically the field in the database
                from which the name must be taken

        Output:
            boolean (if the arguments are valid) or None (if fileType is bad)
        """
        if fileType not in ("arxiv", "inspire", "isbn", "doi", "ads", "scholar"):
            pBLogger.warning(pstr.invalidCheckFileArg)
            return
        return os.path.isfile(self.getFilePath(key, fileType))

    def removeFile(self, key, fileType, fileName=None):
        """Delete a PDF file.

        Parameters:
            key (string): the bibtex key of the entry
            fileType (any string if fileName is not None or string in
                ["arxiv", "inspire", "isbn", "doi", "ads", "scholar"]):
                basically the field in the database
                from which the name must be taken
            fileName (default None): specify the file name instead
                of computing it from the database information

        Output:
            True if successful, False if exception occurred
        """
        if fileName is None:
            fileName = self.getFilePath(key, fileType)
        try:
            os.remove(fileName)
        except OSError:
            pBLogger.exception(pstr.errorRemove % fileName)
            return False
        pBLogger.info(pstr.removed % fileName)
        return True

    def getExisting(self, key, fullPath=False):
        """Obtain the list of existing files for a given entry.

        Parameters:
            key (string): the bibtex key of the entry
            fullPath (boolean, default False):
                return the list with absolute paths

        Output:
            a list, possibly empty
        """
        fileDir = self.getFileDir(key)
        try:
            dircontent = os.listdir(fileDir)
        except Exception:
            return []
        try:
            files = [e for e in dircontent if osp.isfile(osp.join(fileDir, e))]
        except (FileNotFoundError, IOError):
            pBLogger.exception(pstr.errorList)
            return []
        if fullPath:
            files = [osp.join(fileDir, e) for e in files]
        return files

    def printExisting(self, key, fullPath=False):
        """Print the list of existing files for a given entry,
        using self.getExisting to get it.
        Same parameters as self.getExisting.
        """
        pBLogger.info(pstr.listing % (key, self.getFileDir(key)))
        for i, e in enumerate(self.getExisting(key, fullPath=fullPath)):
            pBLogger.info("%2d: %s" % (i, e))

    def printAllExisting(self, entries=None, fullPath=False):
        """Print the complete list of all the existing PDF files
        in the PDF folder, given all the entries in the database
        or a specific subset

        Parameters:
            entries (list, optional): the list of bibtex keys to consider.
                If None, get all the database content
            fullPath (boolean, default False):
                print the list with absolute paths
        """
        iterator = entries
        if entries is None:
            pBDB.bibs.fetchAll(orderBy="firstdate", saveQuery=False, doFetch=False)
            iterator = pBDB.bibs.fetchCursor()
        for e in iterator:
            exist = self.getExisting(e["bibkey"], fullPath=fullPath)
            if len(exist) > 0:
                pBLogger.info("%30s: [%s]" % (e["bibkey"], "] [".join(exist)))

    def removeSparePDFFolders(self):
        """Scans the PDF folder in order to find single unassociated
        directories (their name is not matched by any database entry).
        The unassociated directories are removed without asking any
        additional confirmation. Be careful!
        """
        pBDB.bibs.fetchAll(doFetch=False)
        folders = os.listdir(self.pdfDir)
        for e in pBDB.bibs.fetchCursor():
            k = e["bibkey"]
            cleaned = self.badFName(k)
            if cleaned in folders:
                del folders[folders.index(cleaned)]
        if len(folders) > 0:
            pBLogger.info(pstr.spareFound % (len(folders), folders))
            for f in folders:
                shutil.rmtree(osp.join(self.pdfDir, f))
            pBLogger.info(pstr.doneE)
        else:
            pBLogger.warning(pstr.nothingFound)

    def mergePDFFolders(self, oldkey, newkey):
        """Copy the PDF files that exist in one folder to a new one

        Parameters:
            oldkey: the old bibtex key
            newkey: the new bibtex key
        """
        oldPDFs = self.getExisting(oldkey, fullPath=True)
        outFolder = self.getFileDir(newkey)
        self.createFolder(newkey)
        for o in oldPDFs:
            try:
                shutil.copy2(o, outFolder)
                pBLogger.info(pstr.copied % (o, outFolder))
            except:
                pBLogger.exception(pstr.errorCopy % (o, outFolder))

    def numberOfFiles(self, folder):
        """Get the total number of files inside the given folder

        Parameters:
            folder: the path of the folder to scan

        Output:
            the number of files
        """
        total_number = 0
        for dirpath, dirnames, filenames in os.walk(folder):
            for f in filenames:
                fp = os.path.join(dirpath, f)
                if os.path.isfile(fp):
                    total_number += 1
        return total_number

    def dirSize(self, folder, dirs=True):
        """Get the size of a single directory and its content

        Parameters:
            folder: the path of the folder to scan
            dirs (default True): if True, include the size of folders

        Output:
            the size in bytes
        """
        if dirs:
            try:
                total_size = os.path.getsize(folder)
            except FileNotFoundError:
                pBLogger.exception(pstr.folderMissing % folder)
                os.makedirs(folder)
                return os.path.getsize(folder)
        else:
            total_size = 0
        for dirpath, dirnames, filenames in os.walk(folder):
            if dirs:
                for d in dirnames:
                    total_size += os.path.getsize(os.path.join(dirpath, d))
            for f in filenames:
                fp = os.path.join(dirpath, f)
                if os.path.isfile(fp):
                    total_size += os.path.getsize(fp)
        return total_size

    def getSizeWUnits(self, size, units="MB", fmt="%.2f"):
        """Print a size obtained with `self.dirSize`

        Parameters:
            size:
            units (default "MB"):
            format (default "%.2f"): the format

        Output:
            the string with the size in the appropriate units
        """
        allowedUnits = ["B", "KB", "MB", "GB", "TB", "PB"]
        try:
            exponent = allowedUnits.index(units.upper())
            unitsU = units.upper()
        except ValueError:
            pBLogger.warning(pstr.errorUnits)
            exponent = 2
            unitsU = "MB"
        try:
            sizeWUnits = size / 1024.0 ** (exponent)
        except TypeError:
            pBLogger.warning(pstr.errorSize)
            return "nan"
        try:
            return fmt % sizeWUnits + unitsU
        except (TypeError, ValueError):
            pBLogger.warning(pstr.errorFormat)
            return "%.2f" % sizeWUnits + unitsU


pBPDF = LocalPDF()
